import torch
import numpy as np

from sklearn.metrics.pairwise import cosine_similarity
        
def grad_cosine(grad_1, grad_2):
    cos = np.zeros(len(grad_1))
    for i in range(len(grad_1)):
        cos_arr = grad_1[i] * grad_2[i]
        cos_arr /= np.sqrt(np.sum(grad_1[i] ** 2))
        cos_arr /= np.sqrt(np.sum(grad_2[i] ** 2))
        cos[i] = np.sum(cos_arr)
    return cos

def grad_vs_optimal(grad_list, param_list):
    final_param = param_list[-1]
    cos = []
    for i in range(len(param_list) - 1):
        param = param_list[i]
        grad = grad_list[i]
        ideal_direction = [param[j] - final_param[j] for j in range(len(param))]
        cos.append(grad_cosine(grad, ideal_direction))
    return np.stack(cos)

def plot_grad_flow(named_parameters):
    '''Plots the gradients flowing through different layers in the net during training.
    Can be used for checking for possible gradient vanishing / exploding problems.
    
    Usage: Plug this function in Trainer class after loss.backwards() as 
    "plot_grad_flow(self.model.named_parameters())" to visualize the gradient flow'''
    ave_grads = []
    max_grads= []
    layers = []
    for n, p in named_parameters:
        if(p.requires_grad) and ("bias" not in n):
            layers.append(n)
            ave_grads.append(p.grad.abs().mean())
            max_grads.append(p.grad.abs().max())
    plt.bar(np.arange(len(max_grads)), max_grads, alpha=0.1, lw=1, color="c")
    plt.bar(np.arange(len(max_grads)), ave_grads, alpha=0.1, lw=1, color="b")
    plt.hlines(0, 0, len(ave_grads)+1, lw=2, color="k" )
    plt.xticks(range(0,len(ave_grads), 1), layers, rotation="vertical")
    plt.xlim(left=0, right=len(ave_grads))
    plt.ylim(bottom = -0.001, top=0.02) # zoom in on the lower gradient regions
    plt.xlabel("Layers")
    plt.ylabel("average gradient")
    plt.title("Gradient flow")
    plt.grid(True)
    plt.legend([plt.Line2D([0], [0], color="c", lw=4),
                plt.Line2D([0], [0], color="b", lw=4),
                plt.Line2D([0], [0], color="k", lw=4)], ['max-gradient', 'mean-gradient', 'zero-gradient'])

class GradAnalysis(object):
    def __init__(self, model):
        self.model = model
        self.names = []
        self.params = []
        self.grad = []
        
        self.get_param()
        
    def get_param(self):
        self.params = []
        for n, p in self.model.named_parameters():
            if (p.requires_grad) and ("bias" not in n):
                self.names.append(n)
                self.params.append(p.data.clone().cpu().numpy())
        return self.params
                
    def loss_grad(self, loss):
        # Backward and optimize   
        loss.backward(retain_graph=True)
        self.grad = [
            p.grad.clone().cpu().numpy() 
            for n, p in self.model.named_parameters() 
            if (p.requires_grad) and ("bias" not in n)
        ]
        return self.grad
        
    def clear_grad(self):
        for n, p in self.model.named_parameters():
            if (p.requires_grad) and ("bias" not in n):
                p.grad.data.zero_()



import torch
import numpy as np
from torch.autograd import Variable

from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler

# restore hamiltonian from nonzeros
def restore_h(nonzero, nonzero_loc, dim):
    data_count = nonzero.shape[0]
    H = torch.zeros((data_count, dim, dim), device=nonzero.device).float()
    triangle_shape = (data_count, int(dim * (dim + 1) / 2))
    upper_triangle = torch.zeros(triangle_shape, device=nonzero.device)
    upper_triangle[:, nonzero_loc] = nonzero
    iu1 = np.triu_indices(dim)
    il1 = np.tril_indices(dim)
    H[:, iu1[0], iu1[1]] = upper_triangle
    H_t = torch.transpose(H, dim0=1, dim1=2)
    diag = torch.diagonal(H, offset=0, dim1=1, dim2=2)
    diag_matrix = torch.diag_embed(diag, offset=0, dim1=1, dim2=2)
    H += H_t - diag_matrix
    return H


def getData(n_sites, path_to_data, train_test, dataset='new'):
    if dataset == 'new':
        try:
            # load data from server
            directory = path_to_data + "%d-qubits/100k-instances/single-phase/%s/" % (n_sites, train_test)

            # read the files
            with open(directory + "header.bin", 'rb') as file:
                header = np.fromfile(file, dtype=np.int32)
            with open(directory + "matloc.bin", 'rb') as file:
                matloc = np.fromfile(file, dtype=np.int32)
            with open(directory + "matval.bin", 'rb') as file:
                matval = np.fromfile(file, dtype=np.float64)
            with open(directory + "wavefx.bin", 'rb') as file:
                wavefx = np.fromfile(file, dtype=np.float64)
            with open(directory + "energy.bin", 'rb') as file:
                energy = np.fromfile(file, dtype=np.float64)
            with open(directory + "fields.bin", 'rb') as file:
                fields = np.fromfile(file, dtype=np.float64)
            with open(directory + "szdiag.bin", 'rb') as file:
                szdiag = np.fromfile(file, dtype=np.float64)

            # reshape/format data
            data_count = header[0]
            dim = header[2]
            energy = torch.tensor(energy).float().view(data_count, 1)
            Bx = torch.tensor(fields).float().view(data_count, 1)
            
            szdiag = torch.diag(torch.tensor(szdiag)).float()
            psi = torch.tensor(wavefx).float().view(data_count, dim)
            
            nonzero_H_Ids = matloc.reshape((data_count, -1))
            nonzero_H = torch.tensor(matval).float().view(data_count, -1)

        except Exception as e:
            print(e)

        return Bx, nonzero_H, psi, energy, data_count, nonzero_H_Ids, szdiag
    elif dataset == 'old':
        """
        Compile all the data into arrays for use in NN training ans testing.

        Parameters:
            n_sites (int): The number of sites in your model
            path_to_data (str): The path to the directory where the data is.
            train_test (str): Options 'train' or 'test' depending on desired dataset.

        Returns:
            BxBz (np.ndarray): An array with the ordered Bx, Bz pairs in each row.
            Ham (np.ndarray): An array with the flattened input Hamiltonian in each row.
            targets (np.ndarray): An array with the ground state wavefunction in each row.
            eigenvalues (np.ndarray): An array with the energy eigenvalues for the ground state in each row.
            datapoints (int): The number of datapoints loaded.
        """

        with open('%s%s-site-field-gs-%s.bin' % (path_to_data, n_sites, train_test), 'rb') as f:
            params1 = np.fromfile(f, dtype=np.int32, count=3)
            data1 = np.fromfile(f, dtype=np.float64)
        BxBz = data1[0:params1[0] * params1[1]]
        targets = data1[params1[0] * params1[1]::]
        BxBz = BxBz.reshape((params1[0], params1[1]))
        targets = targets.reshape((params1[0], params1[2]))

        with open('%s%s-site-hamil-gs-%s.bin' % (path_to_data, n_sites, train_test), 'rb') as f:
            params1 = np.fromfile(f, dtype=np.int32, count=3)
            data1 = np.fromfile(f, dtype=np.float64)
        Ham = data1[0:params1[0] * params1[1]]
        Ham = Ham.reshape((params1[0], params1[1]))

        with open('%s%s-site-gs-prior-%s.bin' % (path_to_data, n_sites, train_test), 'rb') as f:
            params1 = np.fromfile(f, dtype=np.int32, count=3)
            data1 = np.fromfile(f, dtype=np.float64)
        eigenvals = data1[params1[0] * (params1[1]**2)::]
        eigenvals = eigenvals.reshape((params1[0], params1[1] * params1[2]))
        eigenvals = eigenvals[:,0]
        eigenvals = eigenvals.reshape((params1[0], 1))
        datapoints = params1[0]

        return [BxBz, Ham, targets, eigenvals, datapoints]


# ------------> data loader start here <----------------
class DatasetLoader(object):
    def __init__(
        self, 
        data_path, 
        n_sites, 
        train_size, 
        val_size, 
        test_size,
        dataset='new'
    ):
        self.dataset = dataset
        
        # store parameters
        self.data_path = data_path
        self.n_site = n_sites
        if dataset == 'old':
            # load training data
            training_set = getData(
                n_sites=n_sites,
                path_to_data=data_path,
                train_test='train',
                dataset=dataset
            )
            
            # load test data
            test_set = getData(
                n_sites=n_sites,
                path_to_data=data_path,
                train_test='test',
                dataset=dataset
            )
        else:
            # load training data
            training_set = getData(
                n_sites=n_sites,
                path_to_data=data_path,
                train_test='training',
                dataset=dataset
            )
            
            # load test data
            test_set = getData(
                n_sites=n_sites,
                path_to_data=data_path,
                train_test='testing',
                dataset=dataset
            )
            val_set = getData(
                n_sites=n_sites,
                path_to_data=data_path,
                train_test='validation',
                dataset=dataset
            )
            
            # interpret training variables
            BxBz_val = val_set[0]
            Ham_val = val_set[1]
            targets_val = val_set[2]
            eigenvals_val = val_set[3]
            
            # phased-locked validation dataset: always positive psi_0
            self.val_X = Ham_val
            self.val_y = np.concatenate((targets_val, eigenvals_val),axis=1)
            self.val_y = torch.tensor(self.val_y).float()
            self.available_val = Ham_val.shape[0]
            self.sign_val = torch.sign(self.val_y[:, 0]).view(-1, 1)
            self.val_y[:, :-1] *= self.sign_val
            
            # szdiag
            self.szdiag_train = training_set[6]
            self.szdiag_val = val_set[6]
            self.szdiag_test = test_set[6]
            
            # original tensors
            self.nonzero_loc = training_set[5][0]
        
        # interpret training variables
        BxBz_train = training_set[0]
        Ham_train = training_set[1]
        targets_train = training_set[2]
        eigenvals_train = training_set[3]
        
        # interpret test variables
        BxBz_test = test_set[0]
        Ham_test = test_set[1]
        targets_test = test_set[2]
        eigenvals_test = test_set[3]

        # phased-locked training dataset: always positive psi_0
        self.train_X = Ham_train
        self.train_y = np.concatenate((targets_train, eigenvals_train),axis=1)
        self.train_y = torch.tensor(self.train_y).float()
        self.available_train = Ham_train.shape[0]
        self.sign_train = torch.sign(self.train_y[:, 0]).view(-1, 1)
        self.train_y[:, :-1] *= self.sign_train
        
        # phased-locked test dataset: always positive psi_0
        self.test_X = Ham_test
        self.test_y = np.concatenate((targets_test, eigenvals_test),axis=1)
        self.test_y = torch.tensor(self.test_y).float()
        self.available_test = Ham_test.shape[0]
        self.sign_test = torch.sign(self.test_y[:, 0]).view(-1, 1)
        self.test_y[:, :-1] *= self.sign_test
       
        # dimensions
        self.x_dim = self.train_X.shape[1]
        self.y_dim = self.train_y.shape[1]
        self.h_height = n_sites * n_sites
        self.h_width = n_sites * n_sites
            
        # sample data
        if dataset == 'new':
            self.sample_data(train_size, val_size, test_size)
        else:
            self.sample_data_old(train_size, val_size, test_size)
        
        # create original copy value to protect from normalization
        self.X_train_origin = torch.tensor(self.X_train).float()
        self.X_val_origin = torch.tensor(self.X_val).float()
        self.X_test_origin = torch.tensor(self.X_test).float()
        self.y_train_origin = torch.tensor(self.y_train).float()
        self.y_val_origin = torch.tensor(self.y_val).float()
        self.y_test_origin = torch.tensor(self.y_test).float()
        
        # normalization
        self.std_scaler_x = None
        self.std_scaler_y = None
        self.X_scale = None
        self.X_mean = None
        self.y_scale = None
        self.y_mean = None
        
        # claim tensors
        self.X_train_tensor = None
        self.y_train_tensor = None
        self.X_val_tensor = None
        self.y_val_tensor = None
        self.X_test_tensor = None
        self.y_test_tensor = None
        self.X_scale_tensor = None
        self.X_mean_tensor = None
        self.y_scale_tensor = None
        self.y_mean_tensor = None

    def normalization(self, x=True, y=True):
        # save parameters
        self.x_norm = x
        self.y_norm = y
        
        # normalize on features
        if x:
            self.std_scaler_x = StandardScaler()
            self.X_train = self.std_scaler_x.fit_transform(self.X_train)
            self.X_val = self.std_scaler_x.transform(self.X_val)
            self.X_test = self.std_scaler_x.transform(self.X_test)
            self.X_scale = np.array(self.std_scaler_x.scale_)
            self.X_mean = np.array(self.std_scaler_x.mean_)
                
        # normalize on features
        if y:
            self.std_scaler_y = StandardScaler()
            self.y_train = self.std_scaler_y.fit_transform(self.y_train)
            self.y_val = self.std_scaler_y.transform(self.y_val)
            self.y_test = self.std_scaler_y.transform(self.y_test)
            self.y_scale = np.array(self.std_scaler_y.scale_)
            self.y_mean = np.array(self.std_scaler_y.mean_)
    
    def torch_tensor(self, device):
        if device is None:
            device = self.device
        else:
            self.device = device
        
        # get the torch tensor of data and send to the designated device
        self.X_train_tensor = torch.tensor(self.X_train).float().to(device)
        self.y_train_tensor = torch.tensor(self.y_train).float().to(device)
        self.X_val_tensor = torch.tensor(self.X_val).float().to(device)
        self.y_val_tensor = torch.tensor(self.y_val).float().to(device)
        self.X_test_tensor = torch.tensor(self.X_test).float().to(device)
        self.y_test_tensor = torch.tensor(self.y_test).float().to(device)

        if self.dataset == 'new':
            self.X_train_origin = restore_h(
                self.X_train_origin.to(device), self.nonzero_loc, self.h_height
            )
            self.X_val_origin = restore_h(
                self.X_val_origin.to(device), self.nonzero_loc, self.h_height
            )
            self.X_test_origin = restore_h(
                self.X_test_origin.to(device), self.nonzero_loc, self.h_height
            )
            self.y_train_origin = self.y_train_origin.to(device)
            self.y_val_origin = self.y_val_origin.to(device)
            self.y_test_origin = self.y_test_origin.to(device)
            
        if self.x_norm:
            self.X_scale_tensor = torch.tensor(self.X_scale).float().to(device)
            self.X_mean_tensor = torch.tensor(self.X_mean).float().to(device)
        if self.y_norm:
            self.y_scale_tensor = torch.tensor(self.y_scale).float().to(device)
            self.y_mean_tensor = torch.tensor(self.y_mean).float().to(device)
                
    def sample_data(self, train_size, val_size, test_size):
        
        # save parameters
        if train_size != 0:
            self.num_train = min(train_size, self.available_train)
        else:
            self.num_train = self.available_train
        if val_size != 0:
            self.num_val = min(val_size, self.available_val)
        else:
            self.num_val = self.available_val
        if test_size != 0:
            self.num_test = min(test_size, self.available_test)
        else:
            self.num_test = self.available_test
        
        self.train_index = np.random.choice(
            self.available_train, 
            self.num_train, 
            replace=False
        )
        
        self.val_index = np.random.choice(
            self.available_val, 
            self.num_val, 
            replace=False
        )
        
        self.test_index = np.random.choice(
            self.available_test, 
            self.num_test, 
            replace=False
        )
        
        self.X_train = self.train_X[self.train_index]
        self.y_train = self.train_y[self.train_index]
        self.X_val = self.val_X[self.val_index]
        self.y_val = self.val_y[self.val_index]
        self.X_test = self.test_X[self.test_index]
        self.y_test = self.test_y[self.test_index]

    def sample_data_old(self, train_size, val_size, test_size):
        
        # save parameters
        self.num_train = train_size
        self.num_val = val_size
        self.num_test = test_size
        
        # get random indexes first
        if self.num_train == 0 and self.num_val == 0:
            self.num_train = self.available_train
            self.X_train = self.train_X.copy()
            self.y_train = self.train_y.copy()
            
            self.num_val = self.available_train
            self.X_val = self.train_X.copy()
            self.y_val = self.train_X.copy()
        else:
            # sampling by indexes
            # if any size set to zero, then use entire dataset
            train_val_index = np.random.choice(
                self.available_train, 
                self.num_train + self.num_val, 
                replace=False
            )
            
            if self.num_train:
                self.train_index = train_val_index[:self.num_train]
                self.X_train = self.train_X[self.train_index]
                self.y_train = self.train_y[self.train_index]

            if self.num_val:    
                self.val_index = train_val_index[self.num_train:]
                self.X_val = self.train_X[self.val_index]
                self.y_val = self.train_y[self.val_index]
        
        if self.num_test == 0:
            self.num_test = self.available_test
            self.X_test = self.test_X
            self.y_test = self.test_y
        else:
            self.test_index = np.random.choice(
                self.available_test, 
                self.num_test, 
                replace=False
            )
            self.X_test = self.test_X[self.test_index]
            self.y_test = self.test_y[self.test_index]            
        


import cmt.DNN
import cmt.config_plots
import cmt.data_loader
import cmt.early_stopping
import cmt.presets
import cmt.training_record_grad
import cmt.training
import cmt.evaluations
import cmt.data_loader
import cmt.parameters
import cmt.gradient
import cmt.loss_functions
import cmt.parse_results



# visualization
import matplotlib.pyplot as plt
import seaborn as sns

# pytorch
import torch
from torch import optim
from torch.autograd import Variable
from torch.utils.data import DataLoader

# sklearn
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler

# scipy
from scipy import signal

# others
import os
import sys
import json
import math
import time
import pickle
import numpy as np
import pandas as pd
from datetime import datetime
from tqdm import tqdm
from fastprogress.fastprogress import master_bar, progress_bar

# disable warnings
import warnings
warnings.filterwarnings('ignore')

# seaborn aesthetic settings
sns.set_style("white")

from distutils.spawn import find_executable


# ====================== self defined scripts below ===================

# overwrite the text rendering using LaTex
if find_executable('latex'):
    from .config_plots import global_settings
    global_settings()

# data loader
from .data_loader import DatasetLoader

# the deep neural networks
from .DNN import DNN

# all loss functions        
from .loss_functions import *

# early stopping
from .early_stopping import EarlyStopping

# parameters
from .parameters import Params
        
# ========================= The Trainer Class =========================
class Trainer(object):
    def __init__(self, master_bar=None, plot=True):
        self.current_hash = ""
        self.master_bar = master_bar
        self.plot_flag = plot
        
    def start(self, param, verbose=False):
        current_str = self.str_now() + str(param.__dict__)
        self.current_hash = hex(hash(current_str))
        
        data_params, nn_params, train_params = dict(), dict(), dict()
        for key, value in param.data_params.items():
            data_params[key] = str(value)
        for key, value in param.nn_params.items():
            nn_params[key] = str(value)
        for key, value in param.train_params.items():
            train_params[key] = str(value)
        statistics = \
        """
==========================================================================
Action: initialize task.
Time: %s
Task Id: %s

io_params:
%s
data_params:
%s
nn_params:
%s
loss_params:
%s
train_params:
%s

    Status: Successful
--------------------------------------------------------------------------

        """ % \
        (
            self.str_now(),
            self.current_hash,
            json.dumps(param.io_params, indent=4),
            json.dumps(data_params, indent=4),
            json.dumps(nn_params, indent=4),
            json.dumps(param.loss_params, indent=4),
            json.dumps(train_params, indent=4)
        )
        if verbose:
            print(statistics)
        
        path = param.io_params['path_out'] + param.name + '_' + self.current_hash
        try:
            os.mkdir(path)
        except OSError:
            print ("Creation of the directory %s failed" % path)
        else:
            out_path = path + '/results/'
            log_path = path + '/logs/'
            self.path_model = path + '/models/'
            self.path_grad = path + '/grads/'
            self.path_state = path + '/states/'
            param.io_params['path_log'] = log_path
            param.io_params['path_out'] = out_path
            os.mkdir(out_path)
            os.mkdir(log_path)
            os.mkdir(self.path_model)
            os.mkdir(self.path_grad)
            os.mkdir(self.path_state)
        
        # load data    
        self.configure_io(param.io_params)
          
        # write log file
        log_file = open(
            self.path_log + 'log_' + self.current_hash + '.txt', 
            "a"
        )
        log_file.write(statistics)
        log_file.close()
        
        data = self.load_data(param.data_params)
        if verbose:
            print('Data loaded.')
        
        # build network
        nn_param = param.nn_params
        nn_param['in_dim'] = data.X_train_tensor.size(1)
        nn_param['out_dim'] = data.y_train_tensor.size(1)
        model = self.build_model(nn_param)
        if verbose:
            print('Model built.')
        
        # training
        train_param = param.train_params
        train_param['loss'] = param.loss_params
        train_param['data'] = data
        train_param['model'] = model
        model, stats = self.train(train_param)
        if verbose:
            print('Training Complete.')
            
        # evaluation
        self.evaluate(model, data, stats, param, name=param.name)
            
    def configure_io(self, io_param):
        # add system environment paths
        if 'env_path' in io_param:
            for path in io_param['env_path']:
                sys.path.append(path)
        self.path_out = io_param['path_out']
        self.path_fig = io_param['path_fig']
        self.path_log = io_param['path_log']
        
    def load_data(self, data_param, verbose=False):
        # load data
        if data_param['dataset'] == 'old':
            data_path = data_param['data_path'] + data_param['phase'] + '/'
        else:
            data_path = data_param['data_path']
        data = DatasetLoader(
            data_path=data_path, 
            n_sites=data_param['n_sites'],
            train_size=data_param['train_size'], 
            val_size=data_param['val_size'], 
            test_size=data_param['test_size']
        )
        
        # normalization
        data.normalization(x=data_param['normalize_input'], y=data_param['normalize_output'])
        
        # convert to torch tensor
        data.torch_tensor(device=data_param['device'])
        
        # print out
        statistics = \
        """
==========================================================================
Action: load data.
Time: %s
Task Id: %s

Training Inputs:    %s
Training Outputs:   %s
Validation Inputs:  %s
Validation Outputs: %s
Test Inputs:  %s
Test Outputs: %s

X Scaler: %s
Y Scaler: %s
Device:   %s
        Status: Successful
--------------------------------------------------------------------------
            
        """ % \
        (
            self.str_now(),
            self.current_hash,
            data.X_train.shape,
            data.y_train.shape,
            data.X_val.shape,
            data.y_val.shape,
            data.X_test.shape,
            data.y_test.shape,
            data.std_scaler_x,
            data.std_scaler_y,
            data_param['device']
        )
        
        if verbose:
            print(statistics)
        
        # write log file
        log_file = open(
                self.path_log + 'log_' + self.current_hash + '.txt', 
                "a"
            )
        log_file.write(statistics)
        log_file.close()
        
        return data
    
    def build_model(self, nn_param, verbose=False):
        # Compile model
        d_in = nn_param['in_dim']
        d_out = nn_param['out_dim']
        depth = nn_param['depth']
        h = nn_param['hidden_size']
        activation = nn_param['activation']
        device = nn_param['device']
        
        # parallelize if possible
        model = DNN(d_in, h, d_out, depth, act=activation)
        
#         # parallelize GPU usage
#         if torch.cuda.is_available():
#             if torch.cuda.device_count() > 4:
#                 model = torch.nn.DataParallel(model)
        
        # using the same intialization
        if os.path.exists('initial_state/init.pt'):
            if verbose: print('Initial state loaded.')
            model.load_state_dict(torch.load('initial_state/init.pt'))
        else:
            if verbose: print('Initial state saved.')
            torch.save(model.state_dict(), 'initial_state/init.pt')
        
        model.to(device)

        # print out
        statistics = \
        """
==========================================================================
Action: build model.
Time: %s
Task Id: %s

Input Dimension:    %d
Output Dimension:   %d
Model Depth:        %d
Hidden State Width: %d

Activation:  %s
Device:      %s

        Status: Successful
--------------------------------------------------------------------------
            
        """ % \
        (
            self.str_now(),
            self.current_hash,
            d_in,
            d_out,
            depth,
            h,
            activation,
            device,
        )
        
        if verbose:
            print(statistics)
        
        # write log file
        log_file = open(
            self.path_log + 'log_' + self.current_hash + '.txt', 
            "a"
        )
        log_file.write(statistics)
        log_file.close()
        return model
    
    def train(self, train_param, verbose=False):
        # unpack
        model = train_param['model']
        data = train_param['data']
        loss_param = train_param['loss']
        loss_param['train_loss'] = train_param['train_loss']
        loss_param['test_loss'] = train_param['test_loss']
        
        # write loss file
        loss_file = open(
            self.path_log + 'loss_' + self.current_hash + '.txt', 
            "a"
        )
        loss_file.write(
            '\n==========================================================================\n'
        )
        
        # get training data loader
        train_batch_size = math.ceil(data.y_train_tensor.size(0) / train_param['num_batch'])
        data_train_loader = DataLoader(
            list(zip(data.X_train_tensor, data.y_train_tensor, data.X_train_origin)), 
            batch_size=train_batch_size, 
            shuffle=True
        )
        
        # get test data loader
        val_batch_size = math.ceil(data.y_val_tensor.size(0) / train_param['num_batch'])
        data_val_loader = DataLoader(
            list(zip(data.X_val_tensor, data.y_val_tensor, data.X_val_origin)), 
            batch_size=val_batch_size, 
            shuffle=True
        )
        
        # get test data loader
        test_batch_size = math.ceil(data.y_test_tensor.size(0) / train_param['num_batch'])
        data_test_loader = DataLoader(
            list(zip(data.X_test_tensor, data.y_test_tensor, data.X_test_origin)), 
            batch_size=test_batch_size, 
            shuffle=True
        )
        
        # Optimizer
        optimizer = train_param['optimizer'](
            model.parameters(), 
            weight_decay=train_param['L2_reg']
        )

        # MSE Loss
        criterion = torch.nn.MSELoss()
        
        # path to save model
        path_to_model = self.path_model + '%s.pt' % self.current_hash
        
        # for early stopping
        if train_param['early_stopping']:
            train_param['early_stopping']['saved_model'] = path_to_model
            early_stopping = EarlyStopping(**train_param['early_stopping'])
            
        # For Recording Losses
        NUMEPOCHS = train_param['epochs']
        train_losses = np.zeros(NUMEPOCHS)
        val_losses = np.zeros(NUMEPOCHS)
        test_losses = np.zeros(NUMEPOCHS)
        train_phy_losses = np.zeros(NUMEPOCHS)
        val_phy_losses = np.zeros(NUMEPOCHS)
        test_phy_losses = np.zeros(NUMEPOCHS)
        train_norm_phy_losses = np.zeros(NUMEPOCHS)
        val_norm_phy_losses = np.zeros(NUMEPOCHS)
        test_norm_phy_losses = np.zeros(NUMEPOCHS)
        train_e_losses = np.zeros(NUMEPOCHS)
        val_e_losses = np.zeros(NUMEPOCHS)
        test_e_losses = np.zeros(NUMEPOCHS)
        train_all_losses = np.zeros(NUMEPOCHS)
        val_all_losses = np.zeros(NUMEPOCHS)
        test_all_losses = np.zeros(NUMEPOCHS)
        lambda_s_train = np.zeros(NUMEPOCHS)
        lambda_e_train = np.zeros(NUMEPOCHS)
        lambda_s_test = np.zeros(NUMEPOCHS)
        lambda_e_test = np.zeros(NUMEPOCHS)
        
        # write log file
        task_timestamp = self.str_now()
        statistics = \
        """
==========================================================================
Action: training model.
Time: %s
Task Id: %s

Number of Epochs:   %d
Train Batch Size:   %d
Test Batch Size:    %d
Optimizer:          %s

Training Loss:      %s
Test Loss:          %s

--------------------------------------------------------------------------
            
        """ % \
        (
            task_timestamp,
            self.current_hash,
            NUMEPOCHS,
            train_batch_size,
            test_batch_size,
            optimizer,
            train_param['train_loss'],
            train_param['test_loss'],
        )
        
        if verbose:
            print(statistics)
        
        # write log file
        log_file = open(
            self.path_log + 'log_' + self.current_hash + '.txt', 
            "a"
        )
        log_file.write(statistics)
        log_file.close()
        loss_file.write(
            'Time: %s\n' % task_timestamp
        )
        loss_file.write(
            'Task Id: %s\n' % self.current_hash
        )

        # training
        loss_file.write('Epoch \t Training \t Test \t\t Loss-Phy \t Loss-E \t Anealing Factor\n')
            
        e_coff = loss_param['lambda_e0']
        s_coff = loss_param['lambda_s']
        
        # the progress bar
        if self.master_bar is not None:
            child_bar = progress_bar(range(NUMEPOCHS), parent=self.master_bar)
            self.master_bar.names = ['train', 'val', 'test']
        else:
            child_bar = range(NUMEPOCHS)

        # record when training started
        train_time = 0
        
        if 'break_loop_early' in train_param:
            if train_param['break_loop_early'] == False:
                stopped_epoch = NUMEPOCHS - 1
        
        # save initial state of the model
        torch.save(model.state_dict(), self.path_state + 'state_0.pt')
        
        for epoch in child_bar:
            model.train()
            start_time = time.time()    # start recording time
            if train_param['train_loss'] != []:
                for batchX, batchY, batchH in data_train_loader:  
                    # Forward pass
                    outputs = model(batchX)
                    
                    # cold-start
                    if loss_param['cold_start']:
                        s_coff += self.cold_start(epoch=epoch, **loss_param['cold_start'])
                        s_coff = max(0, s_coff)

                    lambda_s_train[epoch] = s_coff
                    lambda_e_train[epoch] = e_coff
                    
                    # calculate gradients and save
                    (loss_all,
                     loss_mse,
                     loss_phy, 
                     loss_phy_norm, 
                     loss_e,
                     loss_se) = self.loss_for_grad(
                        data, 
                        train_param['train_loss'],
                        outputs=outputs,
                        e_coff=e_coff,
                        s_coff=s_coff,
                        batchX=batchX, 
                        batchY=batchY,
                        batchH=batchH,
                        norm=loss_param['norm_wf']
                    )
                    grad_all = self.calc_gradient(loss=loss_all, model=model, save_name='train_all_%d.pkl' % (epoch + 1))
                    grad_mse = self.calc_gradient(loss=loss_mse, model=model, save_name='train_mse_%d.pkl' % (epoch + 1))
                    grad_phy = self.calc_gradient(loss=loss_phy, model=model, save_name='train_s_%d.pkl' % (epoch + 1))
                    grad_phy_norm = self.calc_gradient(loss=loss_phy_norm, model=model, save_name='train_train_s_norm_%d.pkl' % (epoch + 1))
                    grad_e = self.calc_gradient(loss=loss_e, model=model, save_name='train_e_%d.pkl' % (epoch + 1))
                    grad_se = self.calc_gradient(loss=loss_se, model=model, save_name='train_se_%d.pkl' % (epoch + 1))
                    
                    # Backward and optimize
                    optimizer.zero_grad()
                    loss = self.loss_func(
                        data, 
                        train_param['train_loss'],
                        outputs=outputs,
                        e_coff=e_coff,
                        s_coff=s_coff,
                        batchX=batchX, 
                        batchY=batchY,
                        batchH=batchH,
                        norm=loss_param['norm_wf']
                    )[0]
                    
                    loss.backward()

                    optimizer.step()
                    
            if train_param['test_loss'] != []:
                for batchX, batchY, batchH in data_test_loader:
                    # Forward pass
                    outputs = model(batchX)
                        
                    # cold-start
                    if loss_param['cold_start']:
                        s_coff += self.cold_start(epoch=epoch, **loss_param['cold_start'])
                        s_coff = max(0, s_coff)

                    lambda_s_test[epoch] = s_coff
                    lambda_e_test[epoch] = e_coff
                    
                    # calculate gradients and save
                    (loss_all,
                     loss_mse,
                     loss_phy, 
                     loss_phy_norm, 
                     loss_e,
                     loss_se) = self.loss_for_grad(
                        data, 
                        train_param['test_loss'],
                        outputs=outputs,
                        e_coff=e_coff,
                        s_coff=s_coff,
                        batchX=batchX, 
                        batchY=None,
                        batchH=batchH,
                        norm=loss_param['norm_wf'],
                    )

                    grad_all = self.calc_gradient(loss=loss_all, model=model, save_name='test_all_%d.pkl' % (epoch + 1))
                    grad_mse = self.calc_gradient(loss=loss_mse, model=model, save_name='test_mse_%d.pkl' % (epoch + 1))
                    grad_phy = self.calc_gradient(loss=loss_phy, model=model, save_name='test_s_%d.pkl' % (epoch + 1))
                    grad_phy_norm = self.calc_gradient(loss=loss_phy_norm, model=model, save_name='test_s_norm_%d.pkl' % (epoch + 1))
                    grad_e = self.calc_gradient(loss=loss_e, model=model, save_name='test_e_%d.pkl' % (epoch + 1))
                    grad_se = self.calc_gradient(loss=loss_se, model=model, save_name='test_se_%d.pkl' % (epoch + 1))
                    
                    loss = self.loss_func(
                        data, 
                        train_param['test_loss'],
                        outputs=outputs,
                        e_coff=e_coff,
                        s_coff=s_coff,
                        batchX=batchX, 
                        batchY=None,
                        batchH=batchH,
                        norm=loss_param['norm_wf'],
                    )[0]

                    # Backward and optimize
                    optimizer.zero_grad()
                    loss.backward()

                    optimizer.step()
            
            end_time = time.time()     # end recording time
            train_time += end_time - start_time   # accumulate training time
            
            # record the loss history
            model.eval()
            
            # save initial state of the model
            torch.save(model.state_dict(), self.path_state + 'state_%d.pt' % (epoch + 1))
            
            train_losses[epoch] = criterion(
                model(data.X_train_tensor), data.y_train_tensor
            ).item()
            val_losses[epoch] = criterion(
                model(data.X_val_tensor), data.y_val_tensor
            ).item()
            test_losses[epoch] = criterion(
                model(data.X_test_tensor), data.y_test_tensor
            ).item()
            
            s_coff = loss_param['lambda_s']    
            # train losses
            (loss, 
            train_phy_losses[epoch], 
            train_norm_phy_losses[epoch], 
            train_e_losses[epoch]) = self.loss_func(
                data, 
                train_param['train_loss'],
                outputs=model(data.X_train_tensor),
                e_coff=e_coff,
                s_coff=s_coff,
                batchX=data.X_train_tensor, 
                batchY=data.y_train_tensor,
                batchH=data.X_train_origin,
                norm=loss_param['norm_wf']
            )
            train_all_losses[epoch] = float(loss)
            
            # val losses
            (loss, 
            val_phy_losses[epoch], 
            val_norm_phy_losses[epoch], 
            val_e_losses[epoch]) = self.loss_func(
                data, 
                train_param['test_loss'], 
                outputs=model(data.X_val_tensor),
                e_coff=e_coff,
                s_coff=s_coff,
                batchX=data.X_val_tensor, 
                batchY=data.y_val_tensor,
                batchH=data.X_val_origin,
                norm=loss_param['norm_wf']
            )
            val_all_losses[epoch] = float(loss)
            
            # test losses
            (loss, 
            test_phy_losses[epoch], 
            test_norm_phy_losses[epoch], 
            test_e_losses[epoch]) = self.loss_func(
                data, 
                train_param['test_loss'],
                outputs=model(data.X_test_tensor),
                e_coff=e_coff,
                s_coff=s_coff,
                batchX=data.X_test_tensor, 
                batchY=data.y_test_tensor,
                batchH=data.X_test_origin,
                norm=loss_param['norm_wf']
            )
            test_all_losses = float(loss)
            
            if epoch % loss_param['anneal_interval'] == 0:
                e_coff *= loss_param['anneal_factor']

            if epoch % train_param['print_interval'] == 0:
                loss_file.write(
                    '%d \t %.8f \t %.8f \t %.8f \t %.8f \t %.8f\n' %
                    (
                        epoch, 
                        train_losses[epoch], 
                        test_losses[epoch], 
                        val_phy_losses[epoch], 
                        val_e_losses[epoch], 
                        e_coff
                    )
                )
            
            # plot loss curve
            if epoch % 1 == 0 and self.master_bar is not None and self.plot_flag:
                y_upper_bound = max(train_losses.max(), val_losses[0].max(), test_losses[0].max())
                x_axis = np.arange(epoch+1) + 1
                graphs = [[x_axis, train_losses[:epoch+1]], [x_axis, val_losses[:epoch+1]], [x_axis, test_losses[:epoch+1]]]
                x_bounds = [0, NUMEPOCHS]
                y_bounds = [0.0, y_upper_bound]
                self.master_bar.update_graph(graphs, x_bounds, y_bounds)
            
            # early stopping
            if train_param['early_stopping']:
                early_stopping(val_losses[epoch], model)
                if early_stopping.early_stop:
                    if 'break_loop_early' in train_param:
                        if train_param['break_loop_early'] == True:
                            break
                        else:
                            stopped_epoch = min(epoch, stopped_epoch)
                    else:
                        break
        
        # record when training stopped and calculate time           
        time_per_epoch = train_time / epoch
        
        if 'break_loop_early' in train_param:
            if train_param['break_loop_early'] == False:
                epoch = stopped_epoch
        
        # print loss in log files
        if verbose and self.master_bar is not None:
            self.master_bar.write('Training stopped at %d/%d.' %(epoch, NUMEPOCHS))
        loss_file.write('Training stopped at %d/%d.' %(epoch, NUMEPOCHS))
        loss_file.write('Training time: %f seconds.' % train_time)
        loss_file.write('\nTraining Complete')
        loss_file.write(
            '\n--------------------------------------------------------------------------\n'
        )
        loss_file.close()
        
        # data frame for losses
        df_loss = pd.DataFrame({
            'train_mse': train_losses,
            'val_mse': train_losses,
            'test_mse': test_losses,
            'train_phy': train_phy_losses,
            'val_phy': val_phy_losses,
            'test_phy': test_phy_losses,
            'train_norm_phy': train_norm_phy_losses,
            'train_e': train_e_losses,
            'val_e': val_e_losses,
            'test_e': test_e_losses,
            'val_norm_phy': val_norm_phy_losses,
            'test_norm_phy': test_norm_phy_losses,
            'train_all': train_all_losses,
            'val_all': val_all_losses,
            'test_all': test_all_losses,
            'lambda_s_train': lambda_s_train,
            'lambda_s_test': lambda_s_test,
            'lambda_e_train': lambda_e_train,
            'lambda_e_test': lambda_e_test
        })
        df_loss.to_csv(self.path_out + "losses_" + self.current_hash + ".csv", index=False)
        
        # training statistics to return
        train_stats = {
            'epoch': epoch,
            'train_time': train_time,
            'time_per_epoch': time_per_epoch
        }
        
        # save or load model
        if train_param['early_stopping']:
            model.load_state_dict(torch.load(path_to_model))
        else:
            torch.save(model.state_dict(), path_to_model)
        return model, train_stats
        
    def str_now(self, date=True, time=True):
        """ Return current time in string. """
        if date and time:
            return datetime.now().strftime("%m-%d-%Y-%H:%M:%S")
        else:
            if date:
                return datetime.now().strftime("%m-%d-%Y")
            if time:
                return datetime.now().strftime("%H:%M:%S")
    
    def loss_for_grad(self, data, loss_list, outputs, e_coff=0.0, s_coff=1.0, batchX=None, batchY=None, batchH=None, norm=False):
        """ 
        Set batchY to None when train on test set. 
        Set batchX to None when only use MSE.
        """
        
        # MSE Loss
        criterion = torch.nn.MSELoss()
        if (batchY is not None) and ('mse_loss' in loss_list):
            loss = criterion(outputs, batchY)
        else:
            loss = 0.0
            
        if batchH is not None:
            origin_input = batchH
        else:
            # inverse transformation for normalization
            if data.std_scaler_x is not None:
                origin_input = inverse_norm(
                    batchX, data.X_scale_tensor, data.X_mean_tensor
                )
            else:
                origin_input = batchX

        if data.std_scaler_y is not None:
            origin_output = inverse_norm(
                outputs, data.y_scale_tensor, data.y_mean_tensor
            )
            origin_y = inverse_norm(
                batchY, data.y_scale_tensor, data.y_mean_tensor
            )
        else:
            origin_output = outputs
            origin_y = batchY

        # physics loss and energy loss
        if 'phy_loss' in loss_list:
            loss_phy = phy_loss(
                origin_output,
                origin_y,
                origin_input,
                norm=norm
            )
        else:
            loss_phy = 0.0
            
        if 'energy_loss' in loss_list:
            loss_e = energy_loss(
                origin_output,
                origin_input
            )
        else:
            loss_e = 0.0
            
        if type(loss_phy) == torch.Tensor or type(loss_e) == torch.Tensor:
            loss += torch.mean(s_coff * loss_phy + e_coff * loss_e)
        
        # for calculating the gradient but not as part of the training
        norm_loss_phy = phy_loss(
            origin_output,
            origin_y,
            origin_input,
            norm=True
        )
        norm_loss_phy = torch.mean(norm_loss_phy)
        loss_phy = phy_loss(
            origin_output,
            origin_y,
            origin_input,
            norm=False
        )
        loss_phy = torch.mean(loss_phy)
        loss_e = energy_loss(
            origin_output,
            origin_input
        )
        loss_e = torch.mean(loss_e)
        loss_se = torch.mean(s_coff * loss_phy + e_coff * loss_e)
        if batchY is not None:
            loss_mse = criterion(outputs, batchY)
        else:
            loss_mse = criterion(outputs, outputs)
        return loss, loss_mse, loss_phy, norm_loss_phy, loss_e, loss_se
    
    def calc_gradient(self, loss, model, save_name=None):
        model.zero_grad()
        loss.backward(retain_graph=True)
        grad = {}
        for name, param in model.named_parameters():
            if param.requires_grad:
                grad[name] = param.grad.clone()
            
        if save_name is not None:
            with open(self.path_grad + save_name, 'wb') as f:
                pickle.dump(grad, f, pickle.HIGHEST_PROTOCOL)
        return grad
    
    def loss_func(self, data, loss_list, outputs, e_coff=0.0, s_coff=1.0, batchX=None, batchY=None, batchH=None, norm=False):
        """ 
        Set batchY to None when train on test set. 
        Set batchX to None when only use MSE.
        """
        
        grad = None
        
        # MSE Loss
        criterion = torch.nn.MSELoss()
        if (batchY is not None) and ('mse_loss' in loss_list):
            loss = criterion(outputs, batchY)
        else:
            loss = 0.0
            
        if batchH is not None:
            origin_input = batchH
        else:
            # inverse transformation for normalization
            if data.std_scaler_x is not None:
                origin_input = inverse_norm(
                    batchX, data.X_scale_tensor, data.X_mean_tensor
                )
            else:
                origin_input = batchX

        if data.std_scaler_y is not None:
            origin_output = inverse_norm(
                outputs, data.y_scale_tensor, data.y_mean_tensor
            )
            origin_y = inverse_norm(
                batchY, data.y_scale_tensor, data.y_mean_tensor
            )
        else:
            origin_output = outputs
            origin_y = batchY

        # physics loss and energy loss
        if 'phy_loss' in loss_list:
            loss_phy = phy_loss(
                origin_output,
                origin_y,
                origin_input,
                norm=norm
            )
        else:
            loss_phy = 0.0
            
        if 'energy_loss' in loss_list:
            loss_e = energy_loss(
                origin_output,
                origin_input
            )
        else:
            loss_e = 0.0
            
        if type(loss_phy) == torch.Tensor or type(loss_e) == torch.Tensor:
            loss += torch.mean(s_coff * loss_phy + e_coff * loss_e)
        
        norm_loss_phy = phy_loss(
            origin_output,
            origin_y,
            origin_input,
            norm=True
        )
        norm_loss_phy = torch.mean(norm_loss_phy).item()
        loss_phy = phy_loss(
            origin_output,
            origin_y,
            origin_input,
            norm=False
        )
        loss_phy = torch.mean(loss_phy).item()
        loss_e = energy_loss(
            origin_output,
            origin_input
        )
        loss_e = torch.mean(loss_e).item()
        return loss, loss_phy, norm_loss_phy, loss_e
    
    def evaluate(self, model, data, stats, param, name='unnamed'):
        """ Evaluate model and save data. """
        
        # eval model
        criterion = torch.nn.MSELoss()
        model.eval()
        
        pred_test = model(data.X_test_tensor).detach()
        true_test = data.y_test_tensor
        pred_val = model(data.X_val_tensor).detach()
        true_val = data.y_val_tensor
        pred_train = model(data.X_train_tensor).detach()
        true_train = data.y_train_tensor
        
        mse_all = criterion(pred_test, true_test)
        mse_wf = criterion(pred_test[:, :-1], true_test[:, :-1])
        mse_e = criterion(pred_test[:, -1], true_test[:, -1])
        
        mse_all_val = criterion(pred_val, true_val)
        mse_wf_val = criterion(pred_val[:, :-1], true_val[:, :-1])
        mse_e_val = criterion(pred_val[:, -1], true_val[:, -1])
        
        mse_all_train = criterion(pred_train, true_train)
        mse_wf_train = criterion(pred_train[:, :-1], true_train[:, :-1])
        mse_e_train = criterion(pred_train[:, -1], true_train[:, -1])
        
        overlap = torch.mean(torch.sum(pred_test * true_test, dim=1) / (
            torch.sqrt(torch.sum(true_test**2, dim=1)) * 
            torch.sqrt(torch.sum(pred_test**2, dim=1))
        ))
        
        overlap_val = torch.mean(torch.sum(pred_val * true_val, dim=1) / (
            torch.sqrt(torch.sum(true_val**2, dim=1)) * 
            torch.sqrt(torch.sum(pred_val**2, dim=1))
        ))
        
        overlap_train = torch.mean(torch.sum(pred_train * true_train, dim=1) / (
            torch.sqrt(torch.sum(true_train**2, dim=1)) * 
            torch.sqrt(torch.sum(pred_train**2, dim=1))
        ))
        
        # write loss file
        model_hash = hex(hash(model))
        path_to_model = self.path_model + '%s.pt' % self.current_hash
        
        # dict to write in result file
        s = {
            'name': name,
            'model_hash': model_hash,
            'model_path': path_to_model,
            'log_path': self.path_log + 'log_' + self.current_hash + '.txt',
            'loss_path': self.path_log + 'loss_' + self.current_hash + '.txt',
            'loss_csv': self.path_out + 'losses_'  + self.current_hash + ".csv",
            'mse_all': mse_all.item(),
            'mse_wave_funtion': mse_wf.item(),
            'mse_energy': mse_e.item(),
            'wave_function_overlap': overlap.item(),
            'val_mse_all': mse_all_val.item(),
            'val_mse_wave_funtion': mse_wf_val.item(),
            'val_mse_energy': mse_e_val.item(),
            'val_wave_function_overlap': overlap_train.item(),
            'train_mse_all': mse_all_train.item(),
            'train_mse_wave_funtion': mse_wf_train.item(),
            'train_mse_energy': mse_e_train.item(),
            'train_wave_function_overlap': overlap_train.item(),
            'param_hash': hex(hash(str(param.__dict__))),
            'params': param.__dict__
        }
        
        # update result with training statistics
        s.update(stats)
        
        # write results
        result_file = open(
            self.path_out + 'result_%s_%s.txt' % (self.str_now(), self.current_hash), 
            "a"
        )
        result_file.write(str(s))
        result_file.close()
    
    def cold_start(self, epoch, mode='sigmoid', lambda_s=0.5, threshold=50, smooth=0.1):
        if mode == 'sigmoid':
            sigmoid = lambda x : 1/(1 + np.exp(-x))
            return np.maximum(lambda_s * (sigmoid((epoch - threshold) * smooth)), 0)
        elif mode == 'inverse-sigmoid':
            sigmoid = lambda x : 1/(1 + np.exp(-x))
            return np.maximum(lambda_s * (1 - (sigmoid((epoch - threshold) * smooth))), 0)
        elif mode == 'quick-start':
            s_coeff = lambda s, alpha, threshold, epoch: s * (1 - (1 + alpha) ** np.minimum(- epoch + threshold, 0))
            return np.maximum(s_coeff(lambda_s, smooth, threshold, epoch), 0)
        elif mode == 'quick-drop':
            s_coeff = lambda s, alpha, threshold, epoch: s * ((1 + alpha) ** np.minimum(- epoch + threshold, 0))
            return np.maximum(s_coeff(lambda_s, smooth, threshold, epoch), 0)


# visualization
import matplotlib.pyplot as plt
import seaborn as sns

# pytorch
import torch
from torch import optim
from torch.autograd import Variable
from torch.utils.data import DataLoader

# sklearn
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler

# scipy
from scipy import signal

# others
import json
import math
import time
import numpy as np
import pandas as pd
from datetime import datetime
from tqdm import tqdm
from fastprogress.fastprogress import master_bar, progress_bar

# disable warnings
import warnings
warnings.filterwarnings('ignore')

from distutils.spawn import find_executable
# ====================== self defined scripts below ===================

# overwrite the text rendering using LaTex
if find_executable('latex'):
    from .config_plots import global_settings
    global_settings()

# data loader
from .data_loader import DatasetLoader

# the deep neural networks
from .DNN import DNN

# all loss functions        
from .loss_functions import *

# early stopping
from .early_stopping import EarlyStopping

# parameters
from .parameters import Params

# ========================= The Trainer Class =========================
class Trainer(object):
    def __init__(self, master_bar=None, plot=True):
        self.current_hash = ""
        self.master_bar = master_bar
        self.plot_flag = plot
        
    def start(self, param, verbose=False):
        current_str = self.str_now() + str(param.__dict__)
        self.current_hash = hex(hash(current_str))
        data_params, nn_params, train_params = dict(), dict(), dict()
        for key, value in param.data_params.items():
            data_params[key] = str(value)
        for key, value in param.nn_params.items():
            nn_params[key] = str(value)
        for key, value in param.train_params.items():
            train_params[key] = str(value)
        statistics = \
        """
==========================================================================
Action: initialize task.
Time: %s
Task Id: %s

io_params:
%s
data_params:
%s
nn_params:
%s
loss_params:
%s
train_params:
%s

    Status: Successful
--------------------------------------------------------------------------

        """ % \
        (
            self.str_now(),
            self.current_hash,
            json.dumps(param.io_params, indent=4),
            json.dumps(data_params, indent=4),
            json.dumps(nn_params, indent=4),
            json.dumps(param.loss_params, indent=4),
            json.dumps(train_params, indent=4)
        )
        if verbose:
            print(statistics)
            
        # load data    
        self.configure_io(param.io_params)
        
        # write log file
        log_file = open(
            self.path_log + 'log_' + self.current_hash + '.txt', 
            "a"
        )
        log_file.write(statistics)
        log_file.close()
        
        data = self.load_data(param.data_params)
        if verbose:
            print('Data loaded.')
        
        # build network
        nn_param = param.nn_params
        nn_param['in_dim'] = data.X_train_tensor.size(1)
        nn_param['out_dim'] = data.y_train_tensor.size(1)
        model = self.build_model(nn_param)
        if verbose:
            print('Model built.')
        
        # training
        train_param = param.train_params
        train_param['loss'] = param.loss_params
        train_param['data'] = data
        train_param['model'] = model
        model, stats = self.train(train_param)
        if verbose:
            print('Training Complete.')
            
        # evaluation
        self.evaluate(model, data, stats, param, name=param.name)
            
    def configure_io(self, io_param):
        # add system environment paths
        if 'env_path' in io_param:
            for path in io_param['env_path']:
                sys.path.append(path)
        self.path_out = io_param['path_out']
        self.path_fig = io_param['path_fig']
        self.path_log = io_param['path_log']
        self.path_model = io_param['path_model']
        
    def load_data(self, data_param, verbose=False):
        # load data
        if data_param['dataset'] == 'old':
            data_path = data_param['data_path'] + data_param['phase'] + '/'
        else:
            data_path = data_param['data_path']
        data = DatasetLoader(
            data_path=data_path, 
            n_sites=data_param['n_sites'],
            train_size=data_param['train_size'], 
            val_size=data_param['val_size'], 
            test_size=data_param['test_size']
        )
        
        # normalization
        data.normalization(x=data_param['normalize_input'], y=data_param['normalize_output'])
        
        # convert to torch tensor
        data.torch_tensor(device=data_param['device'])
        
        # print out
        statistics = \
        """
==========================================================================
Action: load data.
Time: %s
Task Id: %s

Training Inputs:    %s
Training Outputs:   %s
Validation Inputs:  %s
Validation Outputs: %s
Test Inputs:  %s
Test Outputs: %s

X Scaler: %s
Y Scaler: %s
Device:   %s
        Status: Successful
--------------------------------------------------------------------------
            
        """ % \
        (
            self.str_now(),
            self.current_hash,
            data.X_train.shape,
            data.y_train.shape,
            data.X_val.shape,
            data.y_val.shape,
            data.X_test.shape,
            data.y_test.shape,
            data.std_scaler_x,
            data.std_scaler_y,
            data_param['device']
        )
        
        if verbose:
            print(statistics)
        
        # write log file
        log_file = open(
                self.path_log + 'log_' + self.current_hash + '.txt', 
                "a"
            )
        log_file.write(statistics)
        log_file.close()
        
        return data
    
    def build_model(self, nn_param, verbose=False):
        # Compile model
        d_in = nn_param['in_dim']
        d_out = nn_param['out_dim']
        depth = nn_param['depth']
        h = nn_param['hidden_size']
        activation = nn_param['activation']
        device = nn_param['device']
        
        # parallelize if possible
        model = DNN(d_in, h, d_out, depth, act=activation)
#         if torch.cuda.is_available():
#             if torch.cuda.device_count() > 4:
#                 model = torch.nn.DataParallel(model)
        model.to(device)
        
        # print out
        statistics = \
        """
==========================================================================
Action: build model.
Time: %s
Task Id: %s

Input Dimension:    %d
Output Dimension:   %d
Model Depth:        %d
Hidden State Width: %d

Activation:  %s
Device:      %s

        Status: Successful
--------------------------------------------------------------------------
            
        """ % \
        (
            self.str_now(),
            self.current_hash,
            d_in,
            d_out,
            depth,
            h,
            activation,
            device,
        )
        
        if verbose:
            print(statistics)
        
        # write log file
        log_file = open(
            self.path_log + 'log_' + self.current_hash + '.txt', 
            "a"
        )
        log_file.write(statistics)
        log_file.close()
        return model
    
    def train(self, train_param, verbose=False):
        # unpack
        model = train_param['model']
        data = train_param['data']
        loss_param = train_param['loss']
        loss_param['train_loss'] = train_param['train_loss']
        loss_param['test_loss'] = train_param['test_loss']
        
        # write loss file
        loss_file = open(
            self.path_log + 'loss_' + self.current_hash + '.txt', 
            "a"
        )
        loss_file.write(
            '\n==========================================================================\n'
        )
        
        # get training data loader
        train_batch_size = math.ceil(data.y_train_tensor.size(0) / train_param['num_batch'])
        data_train_loader = DataLoader(
            list(zip(data.X_train_tensor, data.y_train_tensor, data.X_train_origin)), 
            batch_size=train_batch_size, 
            shuffle=True
        )
        
        # get test data loader
        val_batch_size = math.ceil(data.y_val_tensor.size(0) / train_param['num_batch'])
        data_val_loader = DataLoader(
            list(zip(data.X_val_tensor, data.y_val_tensor, data.X_val_origin)), 
            batch_size=val_batch_size, 
            shuffle=True
        )
        
        # get test data loader
        test_batch_size = math.ceil(data.y_test_tensor.size(0) / train_param['num_batch'])
        data_test_loader = DataLoader(
            list(zip(data.X_test_tensor, data.y_test_tensor, data.X_test_origin)), 
            batch_size=test_batch_size, 
            shuffle=True
        )
        
        # Optimizer
        optimizer = train_param['optimizer'](
            model.parameters(), 
            weight_decay=train_param['L2_reg']
        )
        
        # MSE Loss
        criterion = torch.nn.MSELoss()
        
        # path to save model
        path_to_model = self.path_model + '%s.pt' % self.current_hash
        
        # for early stopping
        if train_param['early_stopping']:
            train_param['early_stopping']['saved_model'] = path_to_model
            early_stopping = EarlyStopping(**train_param['early_stopping'])
            
        # For Recording Losses
        NUMEPOCHS = train_param['epochs']
        train_losses = np.zeros(NUMEPOCHS)
        val_losses = np.zeros(NUMEPOCHS)
        test_losses = np.zeros(NUMEPOCHS)
        train_phy_losses = np.zeros(NUMEPOCHS)
        val_phy_losses = np.zeros(NUMEPOCHS)
        test_phy_losses = np.zeros(NUMEPOCHS)
        train_norm_phy_losses = np.zeros(NUMEPOCHS)
        val_norm_phy_losses = np.zeros(NUMEPOCHS)
        test_norm_phy_losses = np.zeros(NUMEPOCHS)
        train_e_losses = np.zeros(NUMEPOCHS)
        val_e_losses = np.zeros(NUMEPOCHS)
        test_e_losses = np.zeros(NUMEPOCHS)
        train_all_losses = np.zeros(NUMEPOCHS)
        val_all_losses = np.zeros(NUMEPOCHS)
        test_all_losses = np.zeros(NUMEPOCHS)
        
        # write log file
        task_timestamp = self.str_now()
        statistics = \
        """
==========================================================================
Action: training model.
Time: %s
Task Id: %s

Number of Epochs:   %d
Train Batch Size:   %d
Test Batch Size:    %d
Optimizer:          %s

Training Loss:      %s
Test Loss:          %s

--------------------------------------------------------------------------
            
        """ % \
        (
            task_timestamp,
            self.current_hash,
            NUMEPOCHS,
            train_batch_size,
            test_batch_size,
            optimizer,
            train_param['train_loss'],
            train_param['test_loss'],
        )
        
        if verbose:
            print(statistics)
        
        # write log file
        log_file = open(
            self.path_log + 'log_' + self.current_hash + '.txt', 
            "a"
        )
        log_file.write(statistics)
        log_file.close()
        loss_file.write(
            'Time: %s\n' % task_timestamp
        )
        loss_file.write(
            'Task Id: %s\n' % self.current_hash
        )

        # training
        loss_file.write('Epoch \t Training \t Test \t\t Loss-Phy \t Loss-E \t Anealing Factor\n')
            
        e_coff = loss_param['lambda_e0']
        s_coff = loss_param['lambda_s']
        
        # the progress bar
        if self.master_bar is not None:
            child_bar = progress_bar(range(NUMEPOCHS), parent=self.master_bar)
            self.master_bar.names = ['train', 'val', 'test']
        else:
            child_bar = range(NUMEPOCHS)

        # record when training started
        train_time = 0
        
        if 'break_loop_early' in train_param:
            if train_param['break_loop_early'] == False:
                stopped_epoch = NUMEPOCHS - 1
                
        for epoch in child_bar:
            model.train()
            start_time = time.time()    # start recording time
            if train_param['train_loss'] != []:
                for batchX, batchY, batchH in data_train_loader:  
                    # Forward pass
                    outputs = model(batchX)
                    
                    s_coff = loss_param['lambda_s']
                    
                    # cold-start
                    if loss_param['cold_start']:
                        s_coff += self.cold_start(epoch=epoch, **loss_param['cold_start'])
                        s_coff = max(0, s_coff)


                    loss = self.loss_func(
                        data, 
                        train_param['train_loss'],
                        outputs=outputs,
                        e_coff=e_coff,
                        s_coff=s_coff,
                        batchX=batchX, 
                        batchY=batchY,
                        batchH=batchH,
                        norm=loss_param['norm_wf']
                    )[0]
                    
                    # Backward and optimize
                    optimizer.zero_grad()
                    loss.backward()

                    optimizer.step()
                    
            if train_param['test_loss'] != []:
                for batchX, batchY, batchH in data_test_loader:
                    # Forward pass
                    outputs = model(batchX)

                    s_coff = loss_param['lambda_s']
                    
                    # cold-start
                    if loss_param['cold_start']:
                        s_coff += self.cold_start(epoch=epoch, **loss_param['cold_start'])
                        s_coff = max(0, s_coff)


                    loss = self.loss_func(
                        data, 
                        train_param['test_loss'],
                        outputs=outputs,
                        e_coff=e_coff,
                        s_coff=s_coff,
                        batchX=batchX, 
                        batchY=None,
                        batchH=batchH,
                        norm=loss_param['norm_wf'],
                    )[0]

                    # Backward and optimize
                    optimizer.zero_grad()
                    loss.backward()
                    
                    optimizer.step()
            
            end_time = time.time()     # end recording time
            train_time += end_time - start_time   # accumulate training time
            
            # record the loss history
            model.eval()
            train_losses[epoch] = criterion(
                model(data.X_train_tensor), data.y_train_tensor
            ).item()
            val_losses[epoch] = criterion(
                model(data.X_val_tensor), data.y_val_tensor
            ).item()
            test_losses[epoch] = criterion(
                model(data.X_test_tensor), data.y_test_tensor
            ).item()
            
            
            # train losses
            (loss, 
            train_phy_losses[epoch], 
            train_norm_phy_losses[epoch], 
            train_e_losses[epoch]) = self.loss_func(
                data, 
                train_param['train_loss'],
                outputs=model(data.X_train_tensor),
                e_coff=e_coff,
                s_coff=s_coff,
                batchX=data.X_train_tensor, 
                batchY=data.y_train_tensor,
                batchH=data.X_train_origin,
                norm=loss_param['norm_wf']
            )
            train_all_losses[epoch] = float(loss)
            
            # val losses
            (loss, 
            val_phy_losses[epoch], 
            val_norm_phy_losses[epoch], 
            val_e_losses[epoch]) = self.loss_func(
                data, 
                [], 
                outputs=model(data.X_val_tensor),
                e_coff=e_coff,
                s_coff=s_coff,
                batchX=data.X_val_tensor, 
                batchY=data.y_val_tensor,
                batchH=data.X_val_origin,
                norm=loss_param['norm_wf']
            )
            val_all_losses[epoch] = float(loss)
            
            # test losses
            (loss, 
            test_phy_losses[epoch], 
            test_norm_phy_losses[epoch], 
            test_e_losses[epoch]) = self.loss_func(
                data, 
                train_param['test_loss'],
                outputs=model(data.X_test_tensor),
                e_coff=e_coff,
                s_coff=s_coff,
                batchX=data.X_test_tensor, 
                batchY=data.y_test_tensor,
                batchH=data.X_test_origin,
                norm=loss_param['norm_wf']
            )
            test_all_losses = float(loss)
            
            if epoch % loss_param['anneal_interval'] == 0:
                e_coff *= loss_param['anneal_factor']

            if epoch % train_param['print_interval'] == 0:
                loss_file.write(
                    '%d \t %.8f \t %.8f \t %.8f \t %.8f \t %.8f\n' %
                    (
                        epoch, 
                        train_losses[epoch], 
                        test_losses[epoch], 
                        val_phy_losses[epoch], 
                        val_e_losses[epoch], 
                        e_coff
                    )
                )
            
            # plot loss curve
            if epoch % 1 == 0 and self.master_bar is not None and self.plot_flag:
                y_upper_bound = max(train_losses.max(), val_losses[0].max(), test_losses[0].max())
                x_axis = np.arange(epoch+1) + 1
                graphs = [[x_axis, train_losses[:epoch+1]], [x_axis, val_losses[:epoch+1]], [x_axis, test_losses[:epoch+1]]]
                x_bounds = [0, NUMEPOCHS]
                y_bounds = [0.0, y_upper_bound]
                self.master_bar.update_graph(graphs, x_bounds, y_bounds)
            
            # early stopping
            if train_param['early_stopping']:
                early_stopping(val_losses[epoch], model)
                if early_stopping.early_stop:
                    if 'break_loop_early' in train_param:
                        if train_param['break_loop_early'] == True:
                            break
                        else:
                            stopped_epoch = min(epoch, stopped_epoch)
                    else:
                        break
        
        # record when training stopped and calculate time           
        time_per_epoch = train_time / epoch
        
        if 'break_loop_early' in train_param:
            if train_param['break_loop_early'] == False:
                epoch = stopped_epoch
        
        # print loss in log files
        if verbose and self.master_bar is not None:
            self.master_bar.write('Training stopped at %d/%d.' %(epoch, NUMEPOCHS))
        loss_file.write('Training stopped at %d/%d.' %(epoch, NUMEPOCHS))
        loss_file.write('Training time: %f seconds.' % train_time)
        loss_file.write('\nTraining Complete')
        loss_file.write(
            '\n--------------------------------------------------------------------------\n'
        )
        loss_file.close()
        
        # data frame for losses
        df_loss = pd.DataFrame({
            'train_mse': train_losses,
            'val_mse': train_losses,
            'test_mse': test_losses,
            'train_phy': train_phy_losses,
            'val_phy': val_phy_losses,
            'test_phy': test_phy_losses,
            'train_norm_phy': train_norm_phy_losses,
            'train_e': train_e_losses,
            'val_e': val_e_losses,
            'test_e': test_e_losses,
            'val_norm_phy': val_norm_phy_losses,
            'test_norm_phy': test_norm_phy_losses,
            'train_all': train_all_losses,
            'val_all': val_all_losses,
            'test_all': test_all_losses
        })
        df_loss.to_csv(self.path_out + "losses_" + self.current_hash + ".csv", index=False)
        
        # training statistics to return
        train_stats = {
            'epoch': epoch,
            'train_time': train_time,
            'time_per_epoch': time_per_epoch
        }
        
        # save or load model
        if train_param['early_stopping']:
            model.load_state_dict(torch.load(path_to_model))
        else:
            torch.save(model.state_dict(), path_to_model)
        return model, train_stats
        
    def str_now(self, date=True, time=True):
        """ Return current time in string. """
        if date and time:
            return datetime.now().strftime("%m-%d-%Y-%H:%M:%S")
        else:
            if date:
                return datetime.now().strftime("%m-%d-%Y")
            if time:
                return datetime.now().strftime("%H:%M:%S")
    
    def loss_func(self, data, loss_list, outputs, e_coff=0.0, s_coff=1.0, batchX=None, batchY=None, batchH=None, norm=False):
        """ 
        Set batchY to None when train on test set. 
        Set batchX to None when only use MSE.
        """
        # MSE Loss
        criterion = torch.nn.MSELoss()
        if (batchY is not None) and ('mse_loss' in loss_list):
            loss = criterion(outputs, batchY)
        else:
            loss = 0.0
            
        if batchH is not None:
            origin_input = batchH
        else:
            # inverse transformation for normalization
            if data.std_scaler_x is not None:
                origin_input = inverse_norm(
                    batchX, data.X_scale_tensor, data.X_mean_tensor
                )
            else:
                origin_input = batchX

        if data.std_scaler_y is not None:
            origin_output = inverse_norm(
                outputs, data.y_scale_tensor, data.y_mean_tensor
            )
            origin_y = inverse_norm(
                batchY, data.y_scale_tensor, data.y_mean_tensor
            )
        else:
            origin_output = outputs
            origin_y = batchY

        # physics loss and energy loss
        if 'phy_loss' in loss_list:
            loss_phy = phy_loss(
                origin_output,
                origin_y,
                origin_input,
                norm=norm
            )
        else:
            loss_phy = 0.0
            
        if 'energy_loss' in loss_list:
            loss_e = energy_loss(
                origin_output,
                origin_input
            )
        else:
            loss_e = 0.0
            
        if type(loss_phy) == torch.Tensor or type(loss_e) == torch.Tensor:
            loss += torch.mean(s_coff * loss_phy + e_coff * loss_e)
        
        norm_loss_phy = phy_loss(
            origin_output,
            origin_y,
            origin_input,
            norm=True
        )
        norm_loss_phy = torch.mean(norm_loss_phy).item()
        loss_phy = phy_loss(
            origin_output,
            origin_y,
            origin_input,
            norm=False
        )
        loss_phy = torch.mean(loss_phy).item()
        loss_e = energy_loss(
            origin_output,
            origin_input
        )
        loss_e = torch.mean(loss_e).item()
        return loss, loss_phy, norm_loss_phy, loss_e
    
    def evaluate(self, model, data, stats, param, name='unnamed'):
        """ Evaluate model and save data. """
        
        # eval model
        criterion = torch.nn.MSELoss()
        model.eval()
        
        pred_test = model(data.X_test_tensor).detach()
        true_test = data.y_test_tensor
        pred_val = model(data.X_val_tensor).detach()
        true_val = data.y_val_tensor
        pred_train = model(data.X_train_tensor).detach()
        true_train = data.y_train_tensor
        
        mse_all = criterion(pred_test, true_test)
        mse_wf = criterion(pred_test[:, :-1], true_test[:, :-1])
        mse_e = criterion(pred_test[:, -1], true_test[:, -1])
        
        mse_all_val = criterion(pred_val, true_val)
        mse_wf_val = criterion(pred_val[:, :-1], true_val[:, :-1])
        mse_e_val = criterion(pred_val[:, -1], true_val[:, -1])
        
        mse_all_train = criterion(pred_train, true_train)
        mse_wf_train = criterion(pred_train[:, :-1], true_train[:, :-1])
        mse_e_train = criterion(pred_train[:, -1], true_train[:, -1])
        
        overlap = torch.mean(torch.sum(pred_test * true_test, dim=1) / (
            torch.sqrt(torch.sum(true_test**2, dim=1)) * 
            torch.sqrt(torch.sum(pred_test**2, dim=1))
        ))
        
        overlap_val = torch.mean(torch.sum(pred_val * true_val, dim=1) / (
            torch.sqrt(torch.sum(true_val**2, dim=1)) * 
            torch.sqrt(torch.sum(pred_val**2, dim=1))
        ))
        
        overlap_train = torch.mean(torch.sum(pred_train * true_train, dim=1) / (
            torch.sqrt(torch.sum(true_train**2, dim=1)) * 
            torch.sqrt(torch.sum(pred_train**2, dim=1))
        ))
        
        # write loss file
        model_hash = hex(hash(model))
        path_to_model = 'models/%s.pt' % self.current_hash
        
        # dict to write in result file
        s = {
            'name': name,
            'model_hash': model_hash,
            'model_path': path_to_model,
            'log_path': self.path_log + 'log_' + self.current_hash + '.txt',
            'loss_path': self.path_log + 'loss_' + self.current_hash + '.txt',
            'loss_csv': self.path_out + 'losses_'  + self.current_hash + ".csv",
            'mse_all': mse_all.item(),
            'mse_wave_funtion': mse_wf.item(),
            'mse_energy': mse_e.item(),
            'wave_function_overlap': overlap.item(),
            'val_mse_all': mse_all_val.item(),
            'val_mse_wave_funtion': mse_wf_val.item(),
            'val_mse_energy': mse_e_val.item(),
            'val_wave_function_overlap': overlap_train.item(),
            'train_mse_all': mse_all_train.item(),
            'train_mse_wave_funtion': mse_wf_train.item(),
            'train_mse_energy': mse_e_train.item(),
            'train_wave_function_overlap': overlap_train.item(),
            'param_hash': hex(hash(str(param.__dict__))),
            'params': param.__dict__
        }
        
        # update result with training statistics
        s.update(stats)
        
        # write results
        result_file = open(
            self.path_out + 'result_%s_%s.txt' % (self.str_now(), self.current_hash), 
            "a"
        )
        result_file.write(str(s))
        result_file.close()
    
    def cold_start(self, epoch, mode='sigmoid', lambda_s=0.5, threshold=50, smooth=0.1):
        if mode == 'sigmoid':
            sigmoid = lambda x : 1/(1 + np.exp(-x))
            return np.maximum(lambda_s * (sigmoid((epoch - threshold) * smooth)), 0)
        elif mode == 'inverse-sigmoid':
            sigmoid = lambda x : 1/(1 + np.exp(-x))
            return np.maximum(lambda_s * (1 - (sigmoid((epoch - threshold) * smooth))), 0)
        elif mode == 'quick-start':
            s_coeff = lambda s, alpha, threshold, epoch: s * (1 - (1 + alpha) ** np.minimum(- epoch + threshold, 0))
            return np.maximum(s_coeff(lambda_s, smooth, threshold, epoch), 0)
        elif mode == 'quick-drop':
            s_coeff = lambda s, alpha, threshold, epoch: s * ((1 + alpha) ** np.minimum(- epoch + threshold, 0))
            return np.maximum(s_coeff(lambda_s, smooth, threshold, epoch), 0)
        


import torch
from collections import OrderedDict

# Multi-layer Perceptron
class DNN(torch.nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        output_size,
        depth,
        act=torch.nn.Tanh,
    ):
        super(DNN, self).__init__()
        
        layers = [('input', torch.nn.Linear(input_size, hidden_size))]
        layers.append(('input_activation', act()))
        for i in range(depth): 
            layers.append(
                ('hidden_%d' % i, torch.nn.Linear(hidden_size, hidden_size))
            )
            layers.append(('activation_%d' % i, act()))
        layers.append(('output', torch.nn.Linear(hidden_size, output_size)))

        layerDict = OrderedDict(layers)
        self.layers = torch.nn.Sequential(layerDict)

    def forward(self, x):
        out = self.layers(x)
        return out


### DO NOT CALL THIS WHEN LATEX IS NOT INSTALLED ###

import matplotlib

def global_settings():
    pgf_with_latex = {                      # setup matplotlib to use latex for output
        "pgf.texsystem": "pdflatex",        # change this if using xetex or lautex
        "text.usetex": True,                # use LaTeX to write all text
        "figure.figsize": (9, 5),     # default fig size of 0.9 textwidth
        "figure.dpi": 300,
        "font.family": "serif",
        "font.serif": [],                   # blank entries should cause plots to inherit fonts from the document
        "font.sans-serif": [],
        "font.monospace": [],
        "axes.labelsize": 16,               # LaTeX default is 10pt font.
        'axes.titlesize': 15,
        "font.size": 15,
        "legend.fontsize": 16,               # Make the legend/label fonts a little smaller
        "xtick.labelsize": 16,
        "ytick.labelsize": 16,
#         "pgf.preamble": [
#             r"\usepackage[utf8x]{inputenc}",    # use utf8 fonts becasue your computer can handle it :)
#             r"\usepackage[T1]{fontenc}",        # plots will be generated using this preamble
#             ]
        }

    matplotlib.rcParams.update(pgf_with_latex)


import torch
from math import sqrt

def inverse_norm(batch, scale, mean):
    return batch * scale + mean

def phy_loss(batchPred, batchReal, batchInput, norm=False):
    num_data = batchPred.size(0)
    if batchInput.dim() == 2:
        H_height = int(sqrt(batchInput.size(1)))
        H_width = H_height
    elif batchInput.dim() == 3:
        H_height = batchInput.size(1)
        H_width = batchInput.size(2)
        
    batchC = batchPred[:, 0: H_width]
    batchEg = batchPred[:, -1]
    loss_phy = torch.sum(
            (
                torch.sum(
                    batchInput.view((-1, H_width, H_height)) * batchC.view((-1, 1, H_width)), 
                    dim=2
                ) - batchC * batchEg.view((-1, 1))
            ) ** 2, 
            dim = 1
        )
    if norm:
        loss_phy /= torch.sum(batchC ** 2, dim=1)
    return loss_phy

def energy_loss(batchPred, batchInput):
    H_height = int(sqrt(batchInput.size(1)))
    H_width = H_height
    batchEg = batchPred[:, -1]
    loss_e = torch.exp(batchEg)
    return loss_e

def mean_field_loss(batchPred, batchReal, batchInput, quadratic=False):
    """ To reduce memory usage, fewer local variables are used here. """
    """ 
        E:   batchPred[:, -1].view((-1, 1))
        Psi: batchPred[:, 0: H_width]
        H:   batchInput.view((-1, H_width, H_height))
        
    """
    H_height = int(sqrt(batchInput.size(1)))
    H_width = H_height
    batchC = batchPred[:, 0: H_width]
    batchEg = batchPred[:, -1]
    batchH = batchInput.view((-1, H_width, H_height))
    
    if quadratic: # eq10
        batchSum = batchC.view((-1, 1, H_width)) @ batchH @ batchC.view((-1, H_width, 1))
        left = batchSum.view(-1) / torch.sum(batchC ** 2, dim=1)
    else:   # eq8
        batchSum = torch.sum(
                        batchH * batchC.view((-1, 1, H_width)), 
                        dim=(1, 2)
                    )
        left =  batchSum / torch.sum(batchC, dim=1)
        
    loss_phy = (left - batchEg) ** 2
    return loss_phy



import re

def parse_results(path):
    f = open(path, 'r')
    lines = list(f.readlines())

    d = "".join(lines)
    d = d.replace('\n', '')
    d = d.replace(' ', '')
    d = d.replace(':<', ':"<')
    d = d.replace('>,', '>",')
    d = d.replace(': nan,', ': "nan",')
    d = re.sub(r"('\:)(\w*\w\()", r'\1"\2', d)
    d = re.sub(r"(\))(\},')", r'\1"\2', d)
    d = re.sub(r"(\w\))(,')", r'\1"\2', d)
    f.close()
    return eval(d)


from .parameters import Params

import re
import glob
import os
import json
import numpy as np

class LambdaSearch(object):
    """
      ===============================================
        Registered Models:
      -----------------------------------------------
        Deep Neural Networks        |   DNN
        Shrodinger Loss DNN         |   S-DNN
        Energy Loss S-DNN           |   SE-DNN
        Normalized SE-DNN           |   NSE-DNN
        Extended Training S-DNN     |   S-DNNex
        Extended Training SE-DNN    |   SE-DNNex
        Normalized S-DNNex          |   NS-DNNex
        Normalized SE-DNNex         |   NSE-DNNex
        Label-Free S-DNNex          |   S-DNNex-LF
        Label-Free SE-DNNex         |   SE-DNNex-LF
        Normalized S-DNNex-LF       |   NS-DNNex-LF
        Normalized SE-DNNex-LF      |   NSE-DNNex-LF
      ===============================================
    
    """

    def __init__(self, data_path='datasets/'):
        self.data_path = data_path
    
    def DNN(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The baseline blackbox DNN. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'mse_loss',
        ]
        param.train_params['test_loss'] = []
        return param
    
    def S_DNN(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ Only Shrodinger loss and only on training set. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss', 
        ]
        param.train_params['test_loss'] = []
        return param
    
    def SE_DNN(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ S, E loss only on training set. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss',
            'energy_loss'
        ]
        param.train_params['test_loss'] = []
        return param

    def NSE_DNN(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ S, E loss only on training set. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss',
            'energy_loss'
        ]
        param.train_params['test_loss'] = []
        return param
    
    def S_DNNex(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ Only Shrodinger loss. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss', 
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
        ]
        return param
    
    def SE_DNNex(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The best performed model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss', 
            'energy_loss'
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
            'energy_loss'
        ]
        return param
    
    def NS_DNNex(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ See what will happen without energy loss. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss', 
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
        ]
        return param
    
    def NSE_DNNex(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ Normalized version. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy loss and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])
        
        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'mse_loss',
            'phy_loss', 
            'energy_loss'
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
            'energy_loss'
        ]
        return param
    
    def S_DNNex_LF(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free S-DNNex. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'phy_loss', 
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
        ]
        return param
    
    def SE_DNNex_LF(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'phy_loss',
            'energy_loss'
        ]
        param.train_params['test_loss'] = [
            'phy_loss',
            'energy_loss'
        ]
        return param
        
    def NS_DNNex_LF(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'phy_loss', 
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
        ]
        return param

    def NSE_DNNex_LF(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'phy_loss', 
            'energy_loss'
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
            'energy_loss'
        ]
        return param
    
    
    # --------- alias of LF = LB ------------
    def S_DNNex_LB(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free S-DNNex. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'phy_loss', 
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
        ]
        return param
    
    def SE_DNNex_LB(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = False
        param.train_params['train_loss'] = [
            'phy_loss',
            'energy_loss'
        ]
        param.train_params['test_loss'] = [
            'phy_loss',
            'energy_loss'
        ]
        return param
        
    def NS_DNNex_LB(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'phy_loss', 
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
        ]
        return param

    def NSE_DNNex_LB(self, patience=50, anneal=(0.5, 1.0), lambda_e0=(0.0, 5.0), lambda_s=(0.0, 10.0)):
        
        """ The label-free model. """
        
        param = Params()
        param.data_params['data_path'] = self.data_path
        param.train_params['early_stopping'] = {'patience':patience, 'verbose':False, 'delta':0}
        
        # coefficient for energy and shrodinger loss
        param.loss_params['anneal_interval'] = 10
        param.loss_params['anneal_factor'] = np.random.uniform(low=anneal[0], high=anneal[1])
        param.loss_params['lambda_e0'] = np.random.uniform(low=lambda_e0[0], high=lambda_e0[1])
        param.loss_params['lambda_s'] = np.random.uniform(low=lambda_s[0], high=lambda_s[1])

        # for coefficient of the shrodinger loss
        param.loss_params['cold_start'] = {}    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        param.loss_params['norm_wf'] = True
        param.train_params['train_loss'] = [
            'phy_loss', 
            'energy_loss'
        ]
        param.train_params['test_loss'] = [
            'phy_loss', 
            'energy_loss'
        ]
        return param


# this class is used for dict to object
class ParamObject(object):
    def __init__(self, d):
        self.__dict__ = d    
    
    
class LoadFilter(object):
    """
      ===============================================
        Registered Models:
      -----------------------------------------------
        Deep Neural Networks        |   DNN
        Shrodinger Loss DNN         |   S-DNN
        Energy Loss S-DNN           |   SE-DNN
        Normalized SE-DNN           |   NSE-DNN
        Extended Training S-DNN     |   S-DNNex
        Extended Training SE-DNN    |   SE-DNNex
        Normalized S-DNNex          |   NS-DNNex
        Normalized SE-DNNex         |   NSE-DNNex
        Label-Free S-DNNex          |   S-DNNex-LF
        Label-Free SE-DNNex         |   SE-DNNex-LF
        Normalized S-DNNex-LF       |   NS-DNNex-LF
        Normalized SE-DNNex-LF      |   NSE-DNNex-LF
        cold_start Lambda-S NSE-DNNex |   C-NSE-DNNex
      ===============================================
    
    """
    def __init__(self):
        pass

    def DNN(self, d):
                
        """ The baseline blackbox DNN. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag
    
    def S_DNN(self, d):
        
        """ Only Shrodinger loss and only on training set. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag
    
    def SE_DNN(self, d):
        
        """ S, E loss only on training set. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag

    def NSE_DNN(self, d):
        
        """ S, E loss only on training set. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag
    
    def S_DNNex(self, d):
        
        """ Only Shrodinger loss. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def SE_DNNex(self, d):
        
        """ The best performed model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag
    
    def NS_DNNex(self, d):
        
        """ See what will happen without energy loss. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def NSE_DNNex(self, d):
        
        """ Normalized version. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag
    
    def S_DNNex_LF(self, d):
        
        """ The label-free S-DNNex. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def SE_DNNex_LF(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss',
            'energy_loss'
        ]
        return flag
        
    def NS_DNNex_LF(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag

    def NSE_DNNex_LF(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag
    
    # -------------- alias LF = LB --------------
    def S_DNNex_LB(self, d):
        
        """ The label-free S-DNNex. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def SE_DNNex_LB(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss',
            'energy_loss'
        ]
        return flag
        
    def NS_DNNex_LB(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag

    def NSE_DNNex_LB(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and not param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag
    
    # --------------- cold-start -----------------
        
    def cS_DNN(self, d):
        
        """ Only Shrodinger loss and only on training set. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag
    
    def cSE_DNN(self, d):
        
        """ S, E loss only on training set. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag

    def cNSE_DNN(self, d):
        
        """ S, E loss only on training set. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == []
        return flag
    
    def cS_DNNex(self, d):
        
        """ Only Shrodinger loss. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def cSE_DNNex(self, d):
        
        """ The best performed model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag
    
    def cNS_DNNex(self, d):
        
        """ See what will happen without energy loss. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def cNSE_DNNex(self, d):
        
        """ Normalized version. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True
            
        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'mse_loss',
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag
    
    def cS_DNNex_LF(self, d):
        
        """ The label-free S-DNNex. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag
    
    def cSE_DNNex_LF(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and not param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss',
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss',
            'energy_loss'
        ]
        return flag
        
    def cNS_DNNex_LF(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
        ]
        return flag

    def cNSE_DNNex_LF(self, d):
        
        """ The label-free model. """
        if 'params' in d:
            param = ParamObject(d['params'])

        flag = True

        # for coefficient of the shrodinger loss
        flag = flag and param.loss_params['cold_start']    # cold_start coefficient

        # loss function for NSE-DNNex-LF (label free)
        flag = flag and param.loss_params['norm_wf']
        flag = flag and param.train_params['train_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        flag = flag and param.train_params['test_loss'] == [
            'phy_loss', 
            'energy_loss'
        ]
        return flag


import torch
from torch import optim

import numpy as np

# ========================= The Parameter Class =========================
class Params(object):
    def __init__(
        self, 
        nn_params={}, 
        train_params={},
        io_params={},
        data_params={},
        loss_params={},
        name='default'
    ):
        self.name = name
        
        # default nerual network parameters
        self.nn_params = {
            'hidden_size': 100,
            'depth': 3,               # network depth = 'depth' + 1
            'activation': torch.nn.Tanh,
            'device': torch.device('cuda:0')
        }
        
        # default training parameters
        self.train_params = {
            'epochs': 500,
            'train_loss': [
                'mse_loss', 
                'phy_loss', 
                'energy_loss'
            ],
            'test_loss': [
                'phy_loss', 
                'energy_loss'
            ],
            'num_batch': 1,
            'optimizer': torch.optim.Adamax,
            'L2_reg': 0.0,
            'verbose': False,
            'print_interval': 10,
            'early_stopping': False
        }
        
        # default input/output parameters
        self.io_params = {
            'path_out': 'results/',
            'path_fig': 'figures/',
            'path_log': 'logs/',
            'path_model': 'models/',
            'env_path': [],           # path that will be registered through sys.path
            'use_timestamp': True
        }
        
        # default dataset parameters
        self.data_params = {
            'data_path': '../../datasets/',
            'phase': 'single-phase',  # use 'single-phase', 'dual-phase' or 'multi-phase'
            'n_sites': 4,
            'train_size': 20000,
            'val_size': 2000,
            'test_size': 0,           # set to 0 to use all the test data
            'normalize_input': True,
            'normalize_output': False,
            'device': torch.device('cuda:0'),
            'dataset': 'new'
        }
        
        # default loss function parameters
        self.loss_params = {
            'lambda_s': 1.0,
            'lambda_e0': 0.2,
            'anneal_factor': 0.9,
            'anneal_interval': 50, 
            'norm_wf': False,
            'cold_start': {}
        }

        # update params
        self.nn_params.update(nn_params)
        self.train_params.update(train_params)
        self.io_params.update(io_params)
        self.data_params.update(data_params)
        self.loss_params.update(loss_params)


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='pgml-cmt',
      version='0.2.1',
      description='Navigating the Generalization Landscape of Neural Networks using Physics: A Case Study in Quantum Mechanics',
      url='http://shorturl.at/tCZ05',
      author='Anonymous Author',
      author_email='pgml_cmt@yahoo.com',
      license='MIT',
      packages=['cmt'],
      package_dir={'cmt': 'cmt'},
      long_description=long_description,
      install_requires=[
#           'numpy',
#           'pandas',
#           'tqdm',
#           'matplotlib',
#           'seaborn',
#           'Pillow',
#           'fastprogress',
#           'torch',
#           'torchvision',
#           'scikit-learn',
#           'scipy',
#           'jupyterlab'
      ],
      zip_safe=False)



import torch
import numpy as np
import os
import pickle
import pandas as pd
import glob

def load_grad(grad_path, epoch):
    load = lambda path: pickle.load(open(grad_path + path % epoch, 'rb'))
    grad_dict = {
        'train_mse': load('train_mse_%d.pkl'),
        'train_s_norm': load('train_train_s_norm_%d.pkl'),
        'test_s_norm': load('test_s_norm_%d.pkl'),
        'train_e': load('train_e_%d.pkl'),
        'test_e': load('test_e_%d.pkl'),
        'train_se': load('train_se_%d.pkl'),
        'test_se': load('test_se_%d.pkl'),
        'train_all': load('train_all_%d.pkl'),
        'test_all': load('test_all_%d.pkl')
    }
    return grad_dict

def cosine(w1, w2):
    numerator = torch.sum(w1 * w2)
    denominator = torch.sqrt(torch.sum(w1 ** 2) * torch.sum(w2 ** 2))
    return float(numerator / denominator)

def optimal_direction(state_path, epoch):
    current_state = torch.load(state_path + 'state_%d.pt' % epoch)
    model_path = state_path.replace('/states/', '/models/')+"*"
    model_file = glob.glob(model_path)[0]
    final_state = torch.load(model_file)
    direction = {}
    for key in current_state.keys():
        direction[key] = final_state[key] - current_state[key]
    return direction

def grad_angle(grad_path, state_path, epoch):
    epoch += 1
    
    opt = optimal_direction(state_path, epoch)
    grad = load_grad(grad_path, epoch)
    angles = {}
    for key in opt.keys():
        angle = {}
        for loss in grad.keys():
            # update direction is opposite to the direction of grad
            angle[loss] = -cosine(opt[key], grad[loss][key])
        angles[key] = angle
    return angles

def calc_angles(model, grad_angles):
    param_angles = {}
    for p, value in grad_angles[model][0].items():
        param_angles[p] = {}
        for loss in value.keys():
            param_angles[p][loss] = []
    for angle in grad_angles[model]:
        for p, d in angle.items():
            for loss, value in d.items():
                param_angles[p][loss].append(value)
    for p in param_angles.keys():
        for loss in param_angles[p].keys():
            param_angles[p][loss] = np.array(param_angles[p][loss])
        return param_angles


def projection(direction, grad):
    numerator = torch.sum(direction * grad)
    denominator = torch.sqrt(torch.sum(direction ** 2))
    return float(numerator / denominator)
    
def grad_projection(grad_path, state_path, epoch):
    epoch += 1
    
    opt = optimal_direction(state_path, epoch)
    grad = load_grad(grad_path, epoch)
    projects = {}
    for key in opt.keys():
        project = {}
        for loss in grad.keys():
            # update direction is opposite to the direction of grad
            project[loss] = - projection(opt[key], grad[loss][key])
        projects[key] = project
    return projects

def calc_projections(model, grad_projections):
   
    param_projections = {}
    for p, value in grad_projections[model][0].items():
        param_projections[p] = {}
        for loss in value.keys():
            param_projections[p][loss] = []
    for project in grad_projections[model]:
        for p, d in project.items():
            for loss, value in d.items():
                param_projections[p][loss].append(value)
    for p in param_projections.keys():
        for loss in param_projections[p].keys():
            param_projections[p][loss] = np.array(param_projections[p][loss])
        return param_projections

def concat_params(params):
    flatten = []
    for value in params.values():
        flatten.append(value.view(-1))
    return torch.cat(flatten, dim=0)

def overall_projection(grad_path, state_path, epoch):
    epoch += 1
    
    opt = optimal_direction(state_path, epoch)
    grad = load_grad(grad_path, epoch)
    projects = {}
    for loss in grad.keys():
        projects[loss] = - projection(concat_params(opt), concat_params(grad[loss]))
    return projects

def overall_angle(grad_path, state_path, epoch):
    epoch += 1
    
    opt = optimal_direction(state_path, epoch)
    grad = load_grad(grad_path, epoch)
    results = {}
    for loss in grad.keys():
        results[loss] = - cosine(concat_params(opt), concat_params(grad[loss]))
    return results

def calc_overall_projection():
    overall_projections = {}
    for path in glob.glob('*'):
        name = path.split('_')[0]
        grad_path = path + '/grads/'
        state_path = path + '/states/'
        
        overall_projections[name] = []
        for epoch in range(499):
            project = overall_projection(grad_path, state_path, epoch)
            overall_projections[name].append(project)
            
    for model in overall_projections.keys():
        losses = overall_projections[model][0].keys()
        values = [list(item.values()) for item in overall_projections[model]]
        values = np.stack(values)
        df_projects = pd.DataFrame(columns=losses, data=values)
        overall_projections[model] = df_projects
    return overall_projections

def calc_overall_angle():
    overall_angles = {}
    for path in glob.glob('*'):
        name = path.split('_')[0]
        grad_path = path + '/grads/'
        state_path = path + '/states/'
        
        overall_angles[name] = []
        for epoch in range(499):
            project = overall_angle(grad_path, state_path, epoch)
            overall_angles[name].append(project)
            
    for model in overall_angles.keys():
        losses = overall_angles[model][0].keys()
        values = [list(item.values()) for item in overall_angles[model]]
        values = np.stack(values)
        df_angles = pd.DataFrame(columns=losses, data=values)
        overall_angles[model] = df_angles
    return overall_angles